//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by winhcksp.rc
//
#define IDC_MYICON 2
#define IDD_WINHACK_DIALOG 102
#define IDD_ABOUTBOX 103
#define IDS_APP_TITLE 103
#define IDM_ABOUT 104
#define IDM_EXIT 105
#define IDS_HELLO 106
#define IDI_WINHACK 107
#define IDC_WINHACK 109
#define IDC_SPHONE_DIALOGBAR 111
#define IDC_SPHONE_TEXTDIALOGBAR 112
#define IDR_MAINFRAME 128
#define IDB_TILES 129
#define IDD_TEXT 130
#define IDD_NHTEXT 130
#define IDD_MENU 132
#define IDB_MENU_SEL 133
#define IDB_MENU_UNSEL 134
#define IDD_COMMANDS 136
#define IDD_GETLIN 138
#define IDD_EXTCMD 139
#define IDD_PLAYER_SELECTOR 141
#define IDB_PETMARK 145
#define IDB_MENU_SEL_COUNT 146
#define IDB_KEYPAD 147
#define IDB_MENUBAR 154
#define IDC_TEXT_VIEW 1001
#define IDC_CMD_MOVE_NW 1001
#define IDC_CMD_MOVE_N 1002
#define IDC_MENU_LIST 1003
#define IDC_CMD_MOVE_NE 1003
#define IDC_MENU_TEXT 1004
#define IDC_CMD_MOVE_W 1004
#define IDC_CMD_MOVE_SELF 1005
#define IDC_CMD_MOVE_E 1006
#define IDC_CMD_MOVE_SW 1007
#define IDC_CMD_MOVE_S 1008
#define IDC_CMD_MOVE_SE 1009
#define IDC_CMD_MOVE_UP 1010
#define IDC_CMD_MOVE_DOWN 1011
#define IDC_CMD_5 1012
#define IDC_CMD_A 1013
#define IDC_CMD_B 1014
#define IDC_CMD_C 1015
#define IDC_CMD_D 1016
#define IDC_CMD_E 1017
#define IDC_CMD_F 1018
#define IDC_CMD_G 1019
#define IDC_CMD_H 1020
#define IDC_CMD_I 1021
#define IDC_CMD_J 1022
#define IDC_CMD_K 1023
#define IDC_CMD_L 1024
#define IDC_CMD_M 1025
#define IDC_CMD_N 1026
#define IDC_CMD_O 1027
#define IDC_CMD_P 1028
#define IDC_CMD_Q 1029
#define IDC_CMD_R 1030
#define IDC_CMD_S 1031
#define IDC_CMD_T 1032
#define IDC_CMD_U 1033
#define IDC_CMD_V 1034
#define IDC_CMD_W 1035
#define IDC_CMD_X 1036
#define IDC_CMD_Y 1037
#define IDC_CMD_Z 1038
#define IDC_CMD_AA 1039
#define IDC_CMD_BB 1040
#define IDC_CMD_CC 1041
#define IDC_CMD_DD 1042
#define IDC_CMD_EE 1043
#define IDC_CMD_FF 1044
#define IDC_CMD_GG 1045
#define IDC_CMD_HH 1046
#define IDC_CMD_II 1047
#define IDC_CMD_JJ 1048
#define IDC_CMD_KK 1049
#define IDC_CMD_LL 1050
#define IDC_CMD_MM 1051
#define IDC_CMD_NN 1052
#define IDC_CMD_OO 1053
#define IDC_CMD_PP 1054
#define IDC_CMD_QQ 1055
#define IDC_CMD_RR 1056
#define IDC_CMD_SS 1057
#define IDC_CMD_TT 1058
#define IDC_CMD_UU 1059
#define IDC_CMD_VV 1060
#define IDC_CMD_WW 1061
#define IDC_CMD_XX 1062
#define IDC_CMD_YY 1063
#define IDC_CMD_ZZ 1064
#define IDC_CMD_FIRST 1100
#define IDC_CMD_LAST 1300
#define IDC_GETLIN_EDIT 1309
#define IDC_EXTCMD_LIST 1310
#define IDC_PLSEL_NAME 1314
#define IDC_PLSEL_ROLE_RANDOM 1315
#define IDC_PLSEL_RACE_RANDOM 1318
#define IDC_PLSEL_GENDER_RANDOM 1319
#define IDC_PLSEL_ALIGN_RANDOM 1320
#define IDC_PLSEL_ROLE_LIST 1323
#define IDC_PLSEL_RACE_LIST 1324
#define IDC_PLSEL_ALIGN_LIST 1325
#define IDC_PLSEL_GENDER_LIST 1326
#define IDC_ABOUT_VERSION 1327
#define IDC_TEXT_CONTROL 1331
#define IDC_ABOUT_COPYRIGHT 1332
#define IDC_TEXT_TOGGLE_WRAP 1333
#define IDM_SAVE 32771
#define IDM_HELP_LONG 32772
#define IDM_HELP_COMMANDS 32773
#define IDM_HELP_HISTORY 32774
#define IDM_HELP_INFO_CHAR 32775
#define IDM_HELP_INFO_KEY 32776
#define IDM_HELP_OPTIONS 32777
#define IDM_HELP_OPTIONS_LONG 32778
#define IDM_HELP_EXTCMD 32779
#define IDM_HELP_LICENSE 32780
#define IDM_MAP_TILES 32781
#define IDM_MAP_ASCII4X6 32782
#define IDM_MAP_ASCII6X8 32783
#define IDM_MAP_ASCII8X8 32784
#define IDM_MAP_ASCII16X8 32785
#define IDM_MAP_ASCII7X12 32786
#define IDM_MAP_ASCII8X12 32787
#define IDM_MAP_ASCII16X12 32788
#define IDM_MAP_ASCII12X16 32789
#define IDM_MAP_ASCII10X18 32790
#define IDM_MAP_FIT_TO_SCREEN 32791
#define ID_FILE 32792
#define IDS_CAP_FILE 32793
#define ID_HELP 32794
#define IDS_CAP_HELP 32795
#define IDS_CAP_TEMP 32796
#define ID_MAP 32797
#define IDS_CAP_AMP 32798
#define IDS_CAP_MAP 32799
#define IDM_VIEW_KEYPAD 32800
#define ID_VIEW 32801
#define IDS_CAP_VIEW 32802
#define IDS_CAP_ENTIREMAP 32826
#define IDS_CAP_NORMALMAP 32827
#define IDM_HELP_MENU 32828
#define IDM_VIEW_OPTIONS 32829
#define IDM_DIRECT_COMMAND 32830
#define IDS_TEXT_WRAP 32831
#define IDS_TEXT_UNWRAP 32832

// Next default values for new objects
//
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE 155
#define _APS_NEXT_COMMAND_VALUE 32833
#define _APS_NEXT_CONTROL_VALUE 1334
#define _APS_NEXT_SYMED_VALUE 110
#endif
#endif
