/*
 * NetHack 3.6  mdgrep.h  $NHDT-Date: 1432512785 2015/05/25 00:13:05 $ $NHDT-Branch: master $:$NHDT-Revision: 1.11 $
 * Copyright (c) Kenneth Lorber, Kensington, Maryland, 2008
 * NetHack may be freely redistributed.  See license for details.
 *
 * This file generated by mdgrep.pl version 1.7.
 * DO NOT EDIT!  Your changes will be lost.
 */
static struct grep_var grep_vars[] = { { "0", 0 },
                                       { "1", 1 },
#if defined(ALLDOCS)
                                       { "ALLDOCS", 1 },
#else
                                       { "ALLDOCS", 0 },
#endif
#if defined(AMIGA)
                                       { "AMIGA", 1 },
#else
                                       { "AMIGA", 0 },
#endif
#if defined(AMII_GRAPHICS)
                                       { "AMII_GRAPHICS", 1 },
#else
                                       { "AMII_GRAPHICS", 0 },
#endif
#if defined(ASCIIGRAPH)
                                       { "ASCIIGRAPH", 1 },
#else
                                       { "ASCIIGRAPH", 0 },
#endif
#if defined(BETA)
                                       { "BETA", 1 },
#else
                                       { "BETA", 0 },
#endif
#if defined(BSD_JOB_CONTROL)
                                       { "BSD_JOB_CONTROL", 1 },
#else
                                       { "BSD_JOB_CONTROL", 0 },
#endif
#if defined(CLIPPING)
                                       { "CLIPPING", 1 },
#else
                                       { "CLIPPING", 0 },
#endif
#if defined(COMPRESS)
                                       { "COMPRESS", 1 },
#else
                                       { "COMPRESS", 0 },
#endif
#if defined(DLB)
                                       { "DLB", 1 },
#else
                                       { "DLB", 0 },
#endif
                                       { "FALSE", 0 },
#if defined(GEM_GRAPHICS)
                                       { "GEM_GRAPHICS", 1 },
#else
                                       { "GEM_GRAPHICS", 0 },
#endif
#if defined(GNOME_GRAPHICS)
                                       { "GNOME_GRAPHICS", 1 },
#else
                                       { "GNOME_GRAPHICS", 0 },
#endif
#if defined(HANGUPHANDLING)
                                       { "HANGUPHANDLING", 1 },
#else
                                       { "HANGUPHANDLING", 0 },
#endif
#if defined(INSURANCE)
                                       { "INSURANCE", 1 },
#else
                                       { "INSURANCE", 0 },
#endif
#if defined(LIFE)
                                       { "LIFE", 1 },
#else
                                       { "LIFE", 0 },
#endif
#if defined(MAC)
                                       { "MAC", 1 },
#else
                                       { "MAC", 0 },
#endif
#if defined(MAC_GRAPHICS)
                                       { "MAC_GRAPHICS", 1 },
#else
                                       { "MAC_GRAPHICS", 0 },
#endif
#if defined(MAIL)
                                       { "MAIL", 1 },
#else
                                       { "MAIL", 0 },
#endif
#if defined(MFLOPPY)
                                       { "MFLOPPY", 1 },
#else
                                       { "MFLOPPY", 0 },
#endif
#if defined(MSDOS)
                                       { "MSDOS", 1 },
#else
                                       { "MSDOS", 0 },
#endif
#if defined(MSWIN_GRAPHICS)
                                       { "MSWIN_GRAPHICS", 1 },
#else
                                       { "MSWIN_GRAPHICS", 0 },
#endif
#if defined(NOCWD_ASSUMPTIONS)
                                       { "NOCWD_ASSUMPTIONS", 1 },
#else
                                       { "NOCWD_ASSUMPTIONS", 0 },
#endif
#if defined(NOSAVEONHANGUP)
                                       { "NOSAVEONHANGUP", 1 },
#else
                                       { "NOSAVEONHANGUP", 0 },
#endif
#if defined(OS2)
                                       { "OS2", 1 },
#else
                                       { "OS2", 0 },
#endif
#if defined(POSIX_JOB_CONTROL)
                                       { "POSIX_JOB_CONTROL", 1 },
#else
                                       { "POSIX_JOB_CONTROL", 0 },
#endif
#if defined(QT_GRAPHICS)
                                       { "QT_GRAPHICS", 1 },
#else
                                       { "QT_GRAPHICS", 0 },
#endif
#if defined(RANDOM)
                                       { "RANDOM", 1 },
#else
                                       { "RANDOM", 0 },
#endif
#if defined(SAFERHANGUP)
                                       { "SAFERHANGUP", 1 },
#else
                                       { "SAFERHANGUP", 0 },
#endif
#if defined(SECURE)
                                       { "SECURE", 1 },
#else
                                       { "SECURE", 0 },
#endif
#if defined(SHELL)
                                       { "SHELL", 1 },
#else
                                       { "SHELL", 0 },
#endif
#if defined(SUSPEND)
                                       { "SUSPEND", 1 },
#else
                                       { "SUSPEND", 0 },
#endif
#if defined(TEXTCOLOR)
                                       { "TEXTCOLOR", 1 },
#else
                                       { "TEXTCOLOR", 0 },
#endif
#if defined(TOS)
                                       { "TOS", 1 },
#else
                                       { "TOS", 0 },
#endif
                                       { "TRUE", 1 },
#if defined(TTY_GRAPHICS)
                                       { "TTY_GRAPHICS", 1 },
#else
                                       { "TTY_GRAPHICS", 0 },
#endif
#if defined(UNICODE_DRAWING)
                                       { "UNICODE_DRAWING", 1 },
#else
                                       { "UNICODE_DRAWING", 0 },
#endif
#if defined(UNICODE_PLAYERTEXT)
                                       { "UNICODE_PLAYERTEXT", 1 },
#else
                                       { "UNICODE_PLAYERTEXT", 0 },
#endif
#if defined(UNICODE_WIDEWINPORT)
                                       { "UNICODE_WIDEWINPORT", 1 },
#else
                                       { "UNICODE_WIDEWINPORT", 0 },
#endif
#if defined(UNIX)
                                       { "UNIX", 1 },
#else
                                       { "UNIX", 0 },
#endif
#if defined(USER_SOUNDS)
                                       { "USER_SOUNDS", 1 },
#else
                                       { "USER_SOUNDS", 0 },
#endif
#if defined(USE_TILES)
                                       { "USE_TILES", 1 },
#else
                                       { "USE_TILES", 0 },
#endif
#if defined(VAR_PLAYGROUND)
                                       { "VAR_PLAYGROUND", 1 },
#else
                                       { "VAR_PLAYGROUND", 0 },
#endif
#if defined(VMS)
                                       { "VMS", 1 },
#else
                                       { "VMS", 0 },
#endif
#if defined(WIN32)
                                       { "WIN32", 1 },
#else
                                       { "WIN32", 0 },
#endif
#if defined(WIN32_PLATFORM_HPCPRO)
                                       { "WIN32_PLATFORM_HPCPRO", 1 },
#else
                                       { "WIN32_PLATFORM_HPCPRO", 0 },
#endif
#if defined(WIN32_PLATFORM_WFSP)
                                       { "WIN32_PLATFORM_WFSP", 1 },
#else
                                       { "WIN32_PLATFORM_WFSP", 0 },
#endif
#if defined(WINNT)
                                       { "WINNT", 1 },
#else
                                       { "WINNT", 0 },
#endif
#if defined(WIN_CE)
                                       { "WIN_CE", 1 },
#else
                                       { "WIN_CE", 0 },
#endif
#if defined(WIN_CE_POCKETPC)
                                       { "WIN_CE_POCKETPC", 1 },
#else
                                       { "WIN_CE_POCKETPC", 0 },
#endif
#if defined(WIN_CE_PS2xx)
                                       { "WIN_CE_PS2xx", 1 },
#else
                                       { "WIN_CE_PS2xx", 0 },
#endif
#if defined(WIN_CE_SMARTPHONE)
                                       { "WIN_CE_SMARTPHONE", 1 },
#else
                                       { "WIN_CE_SMARTPHONE", 0 },
#endif
                                       { "WIZARD", 1 },
#if defined(X11_GRAPHICS)
                                       { "X11_GRAPHICS", 1 },
#else
                                       { "X11_GRAPHICS", 0 },
#endif
#if defined(ZEROCOMP)
                                       { "ZEROCOMP", 1 },
#else
                                       { "ZEROCOMP", 0 },
#endif
#if defined(ZLIB_COMP)
                                       { "ZLIB_COMP", 1 },
#else
                                       { "ZLIB_COMP", 0 },
#endif
#if defined(__BEOS__)
                                       { "__BEOS__", 1 },
#else
                                       { "__BEOS__", 0 },
#endif
                                       { 0, 0 } };

/* Command ids */
#define TODO_GREP 1

/* End of file */
