//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "newres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_WINHACK             ICON    DISCARDABLE     "..\\..\\wince\\NETHACK.ICO"

/////////////////////////////////////////////////////////////////////////////
//
// Menubar
//

IDC_WINHACK MENU DISCARDABLE 
BEGIN
    MENUITEM "Entire Map",                  IDM_MAP_FIT_TO_SCREEN
    POPUP "Menu"
    BEGIN
        MENUITEM "Options",                     IDM_VIEW_OPTIONS
        MENUITEM "Keypad",                      IDM_VIEW_KEYPAD
        MENUITEM "Cmd",                         IDM_DIRECT_COMMAND
	MENUITEM SEPARATOR
        MENUITEM "ASCII",                       IDM_MAP_ASCII8X8
        MENUITEM "Tiles",						IDM_MAP_TILES
	MENUITEM SEPARATOR
        MENUITEM "Help",                        IDM_HELP_MENU
        MENUITEM "Save",                        IDM_SAVE
        MENUITEM "Quit",                        IDM_EXIT
    END
END

IDC_SPHONE_DIALOGBAR MENU DISCARDABLE 
BEGIN
    MENUITEM "Done",                        IDOK
    MENUITEM "Cancel",                      IDCANCEL
END

/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_WINHACK ACCELERATORS MOVEABLE PURE 
BEGIN
    "?",            IDM_ABOUT,              ASCII,  ALT
    "/",            IDM_ABOUT,              ASCII,  ALT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOG DISCARDABLE  22, 17, 123, 87
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About"
FONT 8, "System"
BEGIN
    LTEXT           "NetHack",IDC_ABOUT_VERSION,0,2,120,15,SS_NOPREFIX
    LTEXT           "Copyright",IDC_ABOUT_COPYRIGHT,0,20,120,50
    DEFPUSHBUTTON   "OK",IDOK,45,75,30,11,WS_GROUP
END

IDD_MENU DIALOG DISCARDABLE  0, 0, 109, 131
STYLE DS_SETFOREGROUND | DS_CONTROL | WS_POPUP | WS_CLIPSIBLINGS | WS_BORDER
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "List1",IDC_MENU_LIST,"SysListView32",WS_BORDER | 
                    WS_TABSTOP,5,5,100,60
    EDITTEXT        IDC_MENU_TEXT,5,70,100,55,ES_MULTILINE | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL
END

IDD_NHTEXT DIALOG DISCARDABLE  0, 0, 83, 83
STYLE WS_POPUP | WS_BORDER
FONT 8, "System"
BEGIN
    EDITTEXT        IDC_TEXT_CONTROL,5,0,70,75,ES_MULTILINE | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL
END

IDD_EXTCMD DIALOG DISCARDABLE  0, 0, 88, 82
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Extended Commands"
FONT 8, "MS Sans Serif"
BEGIN
    LISTBOX         IDC_EXTCMD_LIST,7,5,75,69,LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
END

IDD_GETLIN DIALOG DISCARDABLE  0, 0, 115, 16
STYLE DS_MODALFRAME | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Question?"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_GETLIN_EDIT,0,0,115,15,ES_AUTOHSCROLL
END

IDD_PLAYER_SELECTOR DIALOG DISCARDABLE  0, 0, 105, 124
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "What are you?"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_PLSEL_NAME,35,0,70,12,ES_AUTOHSCROLL | ES_READONLY
    LTEXT           "Name:",IDC_STATIC,0,0,25,10
    CONTROL         "Random",IDC_PLSEL_ROLE_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,20,40,10
    COMBOBOX        IDC_PLSEL_ROLE_LIST,50,20,50,50,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_RACE_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,45,40,10
    COMBOBOX        IDC_PLSEL_RACE_LIST,50,45,50,45,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_GENDER_RANDOM,"Button",
                    BS_AUTOCHECKBOX | BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,70,
                    40,10
    COMBOBOX        IDC_PLSEL_GENDER_LIST,50,70,50,40,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_ALIGN_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,95,40,10
    COMBOBOX        IDC_PLSEL_ALIGN_LIST,50,95,50,45,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    GROUPBOX        "Role",IDC_STATIC,0,10,105,25,WS_GROUP
    GROUPBOX        "Race",IDC_STATIC,0,35,105,25
    GROUPBOX        "Gender",IDC_STATIC,0,60,105,25
    GROUPBOX        "Alignment",IDC_STATIC,0,85,105,25
    DEFPUSHBUTTON   "Play",IDOK,0,110,55,14
    PUSHBUTTON      "Quit",IDCANCEL,55,110,50,14
END


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""newres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_TILES               BITMAP  DISCARDABLE     "..\\..\\wince\\tiles.bmp"
IDB_MENU_SEL            BITMAP  DISCARDABLE     "..\\..\\wince\\mnsel.bmp"
IDB_MENU_UNSEL          BITMAP  DISCARDABLE     "..\\..\\wince\\mnunsel.bmp"
IDB_PETMARK             BITMAP  DISCARDABLE     "..\\..\\wince\\petmark.bmp"
IDB_MENU_SEL_COUNT      BITMAP  DISCARDABLE     "..\\..\\wince\\mnselcnt.bmp"
IDB_KEYPAD              BITMAP  DISCARDABLE     "..\\..\\wince\\keypad.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_NHTEXT, DIALOG
    BEGIN
        LEFTMARGIN, 5
        RIGHTMARGIN, 76
        TOPMARGIN, 7
    END

    IDD_EXTCMD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 82
        TOPMARGIN, 7
        BOTTOMMARGIN, 75
    END

    IDD_PLAYER_SELECTOR, DIALOG
    BEGIN
        RIGHTMARGIN, 98
        BOTTOMMARGIN, 117
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Data
//

IDC_WINHACK SHMENUBAR DISCARDABLE 
BEGIN
    IDC_WINHACK, 2,
    I_IMAGENONE, IDM_MAP_FIT_TO_SCREEN, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, IDM_MAP_FIT_TO_SCREEN, 0, NOMENU, 
    I_IMAGENONE, ID_VIEW, TBSTATE_ENABLED, TBSTYLE_DROPDOWN | TBSTYLE_AUTOSIZE, IDS_CAP_VIEW, 0, 1,
END

IDC_SPHONE_DIALOGBAR SHMENUBAR DISCARDABLE 
BEGIN
    IDC_SPHONE_DIALOGBAR, 2,
    I_IMAGENONE, IDOK, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, IDOK, 0, NOMENU,
    I_IMAGENONE, IDCANCEL, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, IDCANCEL, 0, 
    NOMENU,
END

IDC_SPHONE_TEXTDIALOGBAR SHMENUBAR DISCARDABLE 
BEGIN
    IDC_SPHONE_DIALOGBAR, 2,
    I_IMAGENONE, IDOK, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, IDOK, 0, NOMENU,
    I_IMAGENONE, IDC_TEXT_TOGGLE_WRAP, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, IDS_TEXT_WRAP, 0, 
    NOMENU,
END


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 3,6,6,0
 PRODUCTVERSION 3,6,6,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x9L
#else
 FILEFLAGS 0x8L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "NetHack 3.6.6 for Smartphone 2002\0"
            VALUE "CompanyName", " \0"
            VALUE "FileDescription", "nethackm\0"
            VALUE "FileVersion", "3, 6, 6, 0\0"
            VALUE "InternalName", "nethackm\0"
            VALUE "LegalCopyright", "Copyright © 1985-2020\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "OriginalFilename", "nethackm.exe\0"
            VALUE "PrivateBuild", "090914\0"
            VALUE "ProductName", "NetHack For Smartphone\0"
            VALUE "ProductVersion", "3, 6, 6, 0\0"
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // !_MAC


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_APP_TITLE           "NetHack"
    IDC_WINHACK             "NETHACK"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDM_MAP_FIT_TO_SCREEN   "Entire Map"
    IDS_CAP_FILE            "File"
    IDS_CAP_HELP            "Help"
    IDS_CAP_MAP             "Map"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDM_VIEW_KEYPAD         "Show/Hide keypad."
    IDS_CAP_VIEW            "Menu"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDOK                    "Done"
    IDCANCEL                "Cancel"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_CAP_ENTIREMAP       "Entire Map"
    IDS_CAP_NORMALMAP       "Normal Map"
    IDM_HELP_MENU           "Help Menu"
	IDS_TEXT_WRAP           "Wrap"
	IDS_TEXT_UNWRAP         "Unwrap"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

