/* resource set indices for GEM_RSC */

#define MENU 0     /* menu */
#define DOABOUT 12 /* STRING in tree MENU */
#define DOQUIT 30  /* STRING in tree MENU */

#define STATUSLINE 1 /* form/dialog */
#define GRABSTATUS 1 /* BOX in tree STATUSLINE */

#define MAPWIN 2    /* form/dialog */
#define MAPBOX 0    /* BOX in tree MAPWIN */
#define MAPCURSOR 1 /* IBOX in tree MAPWIN */

#define ABOUT 3       /* form/dialog */
#define FLYABOUT 0    /* BOX in tree ABOUT */
#define OKABOUT 1     /* BUTTON in tree ABOUT */
#define NETHACKIMG0 3 /* ICON in tree ABOUT */

#define LINES 4     /* form/dialog */
#define FLYLINES 0  /* BOX in tree LINES */
#define QLINE 1     /* BUTTON in tree LINES */
#define LINESLIST 2 /* USERDEF in tree LINES */

#define YNCHOICE 5    /* form/dialog */
#define FLYYNCHOICE 0 /* BOX in tree YNCHOICE */
#define YNPROMPT 1    /* TEXT in tree YNCHOICE */
#define SOMECHARS 2   /* BOX in tree YNCHOICE */
#define YN1 3         /* BUTTON in tree YNCHOICE */
#define YNN 53        /* BUTTON in tree YNCHOICE */
#define ANYCHAR 55    /* BOX in tree YNCHOICE */
#define CHOSENCH 56   /* FBOXTEXT in tree YNCHOICE */
#define COUNT 58      /* FBOXTEXT in tree YNCHOICE */
#define YNOK 59       /* BUTTON in tree YNCHOICE */

#define LINEGET 6    /* form/dialog */
#define FLYLINEGET 0 /* BOX in tree LINEGET */
#define LGPROMPT 1   /* TEXT in tree LINEGET */
#define LGREPLY 2    /* FBOXTEXT in tree LINEGET */
#define QLG 3        /* BUTTON in tree LINEGET */
#define LGOK 4       /* BUTTON in tree LINEGET */

#define DIRECTION 7    /* form/dialog */
#define FLYDIRECTION 0 /* BOX in tree DIRECTION */
#define DIR1 5         /* BOXTEXT in tree DIRECTION */
#define DIR9 21        /* BOXTEXT in tree DIRECTION */
#define DIRDOWN 23     /* BOXTEXT in tree DIRECTION */
#define DIRUP 25       /* BOXTEXT in tree DIRECTION */

#define MSGWIN 8     /* form/dialog */
#define UPMSG 1      /* BOXCHAR in tree MSGWIN */
#define GRABMSGWIN 2 /* BOX in tree MSGWIN */
#define DNMSG 3      /* BOXCHAR in tree MSGWIN */
#define MSGLINES 4   /* USERDEF in tree MSGWIN */

#define NAMEGET 9        /* form/dialog */
#define FLYNAMEGET 0     /* BOX in tree NAMEGET */
#define PLNAME 2         /* FBOXTEXT in tree NAMEGET */
#define NETHACKPICTURE 4 /* BOXTEXT in tree NAMEGET */

#define PAGER 10   /* form/dialog */
#define FLYPAGER 0 /* BOX in tree PAGER */
#define QPAGER 1   /* BUTTON in tree PAGER */

#define NHICON 11 /* form/dialog */
