# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DLevel(object):
    __slots__ = ['_tab']

    # DLevel
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DLevel
    def Dnum(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # DLevel
    def Dlevel(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))

def CreateDLevel(builder, dnum, dlevel):
    builder.Prep(1, 2)
    builder.PrependInt8(dlevel)
    builder.PrependInt8(dnum)
    return builder.Offset()
