//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "newres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_WINHACK             ICON    DISCARDABLE     "..\\..\\wince\\NETHACK.ICO"

/////////////////////////////////////////////////////////////////////////////
//
// Menubar
//

IDC_WINHACK MENU DISCARDABLE 
BEGIN
    POPUP "File"
    BEGIN
        MENUITEM "&Save",                       IDM_SAVE
        MENUITEM SEPARATOR
        MENUITEM "&Quit",                       IDM_EXIT
    END
    POPUP "Map"
    BEGIN
        MENUITEM "&0 - Use Tiles",              IDM_MAP_TILES
        MENUITEM "&1 - ASCII (4x6)",            IDM_MAP_ASCII4X6
        MENUITEM "&2 - ASCII (6x8)",            IDM_MAP_ASCII6X8
        MENUITEM "&3 - ASCII (8x8)",            IDM_MAP_ASCII8X8
        MENUITEM "&4 - ASCII (16x8)",           IDM_MAP_ASCII16X8
        MENUITEM "&5 - ASCII (7x12)",           IDM_MAP_ASCII7X12
        MENUITEM "&6 - ASCII  (8x12)",          IDM_MAP_ASCII8X12
        MENUITEM "&7 - ASCII (16x12)",          IDM_MAP_ASCII16X12
        MENUITEM "&8 - ASCII (12x16)",          IDM_MAP_ASCII12X16
        MENUITEM "&9 - ASCII  (10x18)",         IDM_MAP_ASCII10X18
        MENUITEM SEPARATOR
        MENUITEM "&Fit To Screen",              IDM_MAP_FIT_TO_SCREEN
    END
    POPUP "View"
    BEGIN
        MENUITEM "&Keypad",                     IDM_VIEW_KEYPAD
        MENUITEM "&Options",                    IDM_VIEW_OPTIONS
    END
    POPUP "Help"
    BEGIN
        MENUITEM "&About ...",                  IDM_ABOUT
        MENUITEM "&Long description of the game", IDM_HELP_LONG
        MENUITEM "List of &commands",           IDM_HELP_COMMANDS
        MENUITEM "&History of NetHack",         IDM_HELP_HISTORY
        MENUITEM "&Info on a character",        IDM_HELP_INFO_CHAR
        MENUITEM "Info on what a given &key does", IDM_HELP_INFO_KEY
        MENUITEM "List of game &options",       IDM_HELP_OPTIONS
        MENUITEM "&Longer list of game options", IDM_HELP_OPTIONS_LONG
        MENUITEM "List of e&xtended commands",  IDM_HELP_EXTCMD
        MENUITEM "The &NetHack license",        IDM_HELP_LICENSE
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_WINHACK ACCELERATORS MOVEABLE PURE 
BEGIN
    "?",            IDM_ABOUT,              ASCII,  ALT
    "/",            IDM_ABOUT,              ASCII,  ALT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOG DISCARDABLE  22, 17, 123, 87
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About"
FONT 8, "System"
BEGIN
    LTEXT           "NetHack",IDC_ABOUT_VERSION,0,2,120,15,SS_NOPREFIX
    LTEXT           "Copyright",IDC_ABOUT_COPYRIGHT,0,20,120,50
    DEFPUSHBUTTON   "OK",IDOK,45,75,30,11,WS_GROUP
END

IDD_MENU DIALOG DISCARDABLE  0, 0, 109, 153
STYLE WS_CHILD | WS_CLIPSIBLINGS | WS_BORDER
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,5,130,50,14,BS_NOTIFY
    PUSHBUTTON      "Cancel",IDCANCEL,55,130,50,14,BS_NOTIFY
    CONTROL         "List1",IDC_MENU_LIST,"SysListView32",WS_BORDER | 
                    WS_TABSTOP,5,5,100,60
    EDITTEXT        IDC_MENU_TEXT,5,70,100,55,ES_MULTILINE | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL
END

IDD_GETLIN DIALOG DISCARDABLE  0, 0, 115, 30
STYLE DS_MODALFRAME | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Question?"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,0,16,55,14
    PUSHBUTTON      "Cancel",IDCANCEL,55,16,60,14
    EDITTEXT        IDC_GETLIN_EDIT,0,0,115,15,ES_AUTOHSCROLL
END

IDD_PLAYER_SELECTOR DIALOG DISCARDABLE  0, 0, 105, 124
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "What are you?"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_PLSEL_NAME,35,0,70,12,ES_AUTOHSCROLL | ES_READONLY
    LTEXT           "Name:",IDC_STATIC,0,0,25,10
    CONTROL         "Random",IDC_PLSEL_ROLE_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,20,40,10
    COMBOBOX        IDC_PLSEL_ROLE_LIST,50,20,50,50,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_RACE_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,45,40,10
    COMBOBOX        IDC_PLSEL_RACE_LIST,50,45,50,45,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_GENDER_RANDOM,"Button",
                    BS_AUTOCHECKBOX | BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,70,
                    40,10
    COMBOBOX        IDC_PLSEL_GENDER_LIST,50,70,50,40,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    CONTROL         "Random",IDC_PLSEL_ALIGN_RANDOM,"Button",BS_AUTOCHECKBOX | 
                    BS_NOTIFY | WS_GROUP | WS_TABSTOP,5,95,40,10
    COMBOBOX        IDC_PLSEL_ALIGN_LIST,50,95,50,45,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_GROUP | WS_TABSTOP
    GROUPBOX        "Role",IDC_STATIC,0,10,105,25,WS_GROUP
    GROUPBOX        "Race",IDC_STATIC,0,35,105,25
    GROUPBOX        "Gender",IDC_STATIC,0,60,105,25
    GROUPBOX        "Alignment",IDC_STATIC,0,85,105,25
    DEFPUSHBUTTON   "Play",IDOK,0,110,55,14
    PUSHBUTTON      "Quit",IDCANCEL,55,110,50,14
END

IDD_NHTEXT DIALOG DISCARDABLE  0, 0, 100, 97
STYLE WS_CHILD | WS_BORDER
FONT 8, "System"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,0,80,50,14
	PUSHBUTTON      "Wrap",IDC_TEXT_TOGGLE_WRAP,50,80,50,14
    EDITTEXT        IDC_TEXT_CONTROL,5,0,70,75,ES_MULTILINE | ES_READONLY | 
                    WS_VSCROLL | WS_HSCROLL
END

IDD_EXTCMD DIALOG DISCARDABLE  0, 0, 109, 114
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Extended Commands"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,75,5,30,14
    PUSHBUTTON      "Cancel",IDCANCEL,75,20,30,14
    LISTBOX         IDC_EXTCMD_LIST,5,5,65,105,LBS_NOINTEGRALHEIGHT | 
                    WS_VSCROLL | WS_TABSTOP
END


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""newres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_TILES               BITMAP  DISCARDABLE     "..\\..\\wince\\tiles.bmp"
IDB_MENU_SEL            BITMAP  DISCARDABLE     "..\\..\\wince\\mnsel.bmp"
IDB_MENU_UNSEL          BITMAP  DISCARDABLE     "..\\..\\wince\\mnunsel.bmp"
IDB_PETMARK             BITMAP  DISCARDABLE     "..\\..\\wince\\petmark.bmp"
IDB_MENU_SEL_COUNT      BITMAP  DISCARDABLE     "..\\..\\wince\\mnselcnt.bmp"
IDB_KEYPAD              BITMAP  DISCARDABLE     "..\\..\\wince\\keypad.bmp"
IDB_MENUBAR             BITMAP  DISCARDABLE     "..\\..\\wince\\menubar.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_PLAYER_SELECTOR, DIALOG
    BEGIN
        RIGHTMARGIN, 98
        BOTTOMMARGIN, 117
    END

    IDD_NHTEXT, DIALOG
    BEGIN
        LEFTMARGIN, 5
        RIGHTMARGIN, 76
        TOPMARGIN, 7
        BOTTOMMARGIN, 94
    END

    IDD_EXTCMD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 102
        TOPMARGIN, 7
        BOTTOMMARGIN, 107
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Data
//

IDC_WINHACK SHMENUBAR DISCARDABLE 
BEGIN
    IDC_WINHACK, 6,
    I_IMAGENONE, ID_FILE, TBSTATE_ENABLED, 
    TBSTYLE_DROPDOWN | TBSTYLE_AUTOSIZE, IDS_CAP_FILE, 0, 0,
    I_IMAGENONE, ID_MAP, TBSTATE_ENABLED, 
    TBSTYLE_DROPDOWN | TBSTYLE_AUTOSIZE, IDS_CAP_MAP, 0, 1,
    I_IMAGENONE, ID_VIEW, TBSTATE_ENABLED, 
    TBSTYLE_DROPDOWN | TBSTYLE_AUTOSIZE, IDS_CAP_VIEW, 0, 2,
    I_IMAGENONE, ID_HELP, TBSTATE_ENABLED, 
    TBSTYLE_DROPDOWN | TBSTYLE_AUTOSIZE, IDS_CAP_HELP, 0, 3,
    0, IDM_MAP_FIT_TO_SCREEN, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, 0, 
    IDM_MAP_FIT_TO_SCREEN, NOMENU,
    1, IDM_VIEW_KEYPAD, TBSTATE_ENABLED, TBSTYLE_AUTOSIZE, 0, 
    IDM_VIEW_KEYPAD, NOMENU,
END


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 3,6,6,0
 PRODUCTVERSION 3,6,6,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x9L
#else
 FILEFLAGS 0x8L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "NetHack 3.6.6 for Windows CE\0"
            VALUE "CompanyName", " \0"
            VALUE "FileDescription", "nethackm\0"
            VALUE "FileVersion", "3, 6, 6, 0\0"
            VALUE "InternalName", "nethackm\0"
            VALUE "LegalCopyright", "Copyright © 1985-2020\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "OriginalFilename", "nethackm.exe\0"
            VALUE "PrivateBuild", "090914\0"
            VALUE "ProductName", "NetHack\0"
            VALUE "ProductVersion", "3, 6, 6, 0\0"
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // !_MAC


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_APP_TITLE           "NetHack"
    IDC_WINHACK             "NETHACK"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDM_MAP_FIT_TO_SCREEN   "Fit to Screen"
    IDS_CAP_FILE            "File"
    IDS_CAP_HELP            "Help"
    IDS_CAP_MAP             "Map"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDM_VIEW_KEYPAD         "Show/Hide keypad."
    IDS_CAP_VIEW            "View"
	IDS_TEXT_WRAP           "Wrap"
	IDS_TEXT_UNWRAP         "Unwrap"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDOK                    "Done"
    IDCANCEL                "Cancel"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

