/* NetHack 3.6	gnomeprv.h	$NHDT-Date: 1432512806 2015/05/25 00:13:26 $  $NHDT-Branch: master $:$NHDT-Revision: 1.9 $ */
/* Copyright (C) 1998 by Erik Andersen <andersee@debian.org> */
/* NetHack may be freely redistributed.  See license for details. */

#ifndef GnomeHack_h
#define GnomeHack_h

/* These are the base nethack include files */
#include "hack.h"
#include "dlb.h"
#include "patchlevel.h"

#include "winGnome.h"

#endif /* GnomeHack_h */
