/* NetHack 3.6	gnstatus.h	$NHDT-Date: 1432512805 2015/05/25 00:13:25 $  $NHDT-Branch: master $:$NHDT-Revision: 1.8 $ */
/* Copyright (C) 1998 by Erik Andersen <andersee@debian.org> */
/* NetHack may be freely redistributed.  See license for details. */

#ifndef GnomeHackStatusWindow_h
#define GnomeHackStatusWindow_h

#include <gnome.h>
#include "config.h"
#include "global.h"

GtkWidget *ghack_init_status_window();

#endif /* GnomeHackStatusWindow_h */
