/*
 * Use this stub to insert a needed pragma prior
 * to compiling $(PDCWINCON)\pdcscrn.c in the
 * PDCurses distribution.
 *
 * We don't want to ask NetHack builders to
 * modify the PDCurses distribution in any way,
 * we only ask that they have the original intact
 * PDCurses source tree available during the build.
 *
 */

#pragma warning(disable : 4996)
#include "pdcscrn.c"

