# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class You(object):
    __slots__ = ['_tab']

    # You
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # You
    def Ux(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # You
    def Uy(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))
    # You
    def Ux0(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # You
    def Uy0(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3))
    # You
    def Uz(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 4)
        return obj

    # You
    def Uz0(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 6)
        return obj

    # You
    def Uhunger(self): return self._tab.Get(flatbuffers.number_types.Int16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateYou(builder, ux, uy, ux0, uy0, uz_dnum, uz_dlevel, uz0_dnum, uz0_dlevel, uhunger):
    builder.Prep(2, 10)
    builder.PrependInt16(uhunger)
    builder.Prep(1, 2)
    builder.PrependInt8(uz0_dlevel)
    builder.PrependInt8(uz0_dnum)
    builder.Prep(1, 2)
    builder.PrependInt8(uz_dlevel)
    builder.PrependInt8(uz_dnum)
    builder.PrependInt8(uy0)
    builder.PrependInt8(ux0)
    builder.PrependInt8(uy)
    builder.PrependInt8(ux)
    return builder.Offset()
