# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Seeds(object):
    __slots__ = ['_tab']

    # Seeds
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Seeds
    def Core(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Seeds
    def Disp(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateSeeds(builder, core, disp):
    builder.Prep(8, 16)
    builder.PrependUint64(disp)
    builder.PrependUint64(core)
    return builder.Offset()
